@@ -1,6 +1,6 @@
 /*
  * icheck.c --- given a list of blocks, generate a list of inodes
- * 
+ *
  * Copyright (C) 1994 Theodore Ts'o.  This file may be redistributed
  * under the terms of the GNU Public License.
  */
@@ -48,7 +48,7 @@ static int icheck_proc(ext2_filsys fs EXT2FS_ATTR((unused)),
 	}
 	if (!bw->blocks_left)
 		return BLOCK_ABORT;
-	
+
 	return 0;
 }
 
@@ -62,7 +62,7 @@ void do_icheck(int argc, char **argv)
 	struct ext2_inode	inode;
 	errcode_t		retval;
 	char			*block_buf;
-	
+
 	if (argc < 2) {
 		com_err(argv[0], 0, "Usage: icheck <block number> ...");
 		return;
@@ -86,7 +86,7 @@ void do_icheck(int argc, char **argv)
 
 	for (i=1; i < argc; i++) {
 		if (strtoblk(argv[0], argv[i], &bw.barray[i-1].blk))
-			return;
+			goto error_out;
 	}
 
 	bw.num_blocks = bw.blocks_left = argc-1;
@@ -104,7 +104,7 @@ void do_icheck(int argc, char **argv)
 		com_err("icheck", retval, "while starting inode scan");
 		goto error_out;
 	}
-	
+
 	while (ino) {
 		if (!inode.i_links_count)
 			goto next;
@@ -127,7 +127,8 @@ void do_icheck(int argc, char **argv)
 		if (inode.i_dtime)
 			goto next;
 
-		retval = ext2fs_block_iterate2(current_fs, ino, 0, block_buf,
+		retval = ext2fs_block_iterate2(current_fs, ino,
+					       BLOCK_FLAG_READ_ONLY, block_buf,
 					       icheck_proc, &bw);
 		if (retval) {
 			com_err("icheck", retval,
@@ -160,7 +161,8 @@ void do_icheck(int argc, char **argv)
 
 error_out:
 	free(bw.barray);
-	free(block_buf);
+	if (block_buf)
+		free(block_buf);
 	if (scan)
 		ext2fs_close_inode_scan(scan);
 	return;
