@@ -7,14 +7,14 @@
  * This file may be redistributed under the terms of the GNU Public
  * License.
  * %End-Header%
- * 
+ *
  * Pass #3 assures that all directories are connected to the
  * filesystem tree, using the following algorithm:
  *
  * First, the root directory is checked to make sure it exists; if
  * not, e2fsck will offer to create a new one.  It is then marked as
  * "done".
- * 
+ *
  * Then, pass3 interates over all directory inodes; for each directory
  * it attempts to trace up the filesystem tree, using dirinfo.parent
  * until it reaches a directory which has been marked "done".  If it
@@ -24,7 +24,7 @@
  * twice, then it has detected a filesystem loop, and it will again
  * offer to reconnect the directory to /lost+found in to break the
  * filesystem loop.
- * 
+ *
  * Pass 3 also contains the subroutine, e2fsck_reconnect_file() to
  * reconnect inodes to /lost+found; this subroutine is also used by
  * pass 4.  e2fsck_reconnect_file() calls get_lost_and_found(), which
@@ -42,17 +42,17 @@
 #include "problem.h"
 
 static void check_root(e2fsck_t ctx);
-static int check_directory(e2fsck_t ctx, struct dir_info *dir,
+static int check_directory(e2fsck_t ctx, ext2_ino_t ino,
 			   struct problem_context *pctx);
-static void fix_dotdot(e2fsck_t ctx, struct dir_info *dir, ext2_ino_t parent);
+static void fix_dotdot(e2fsck_t ctx, ext2_ino_t ino, ext2_ino_t parent);
 
 static ext2fs_inode_bitmap inode_loop_detect = 0;
 static ext2fs_inode_bitmap inode_done_map = 0;
-	
+
 void e2fsck_pass3(e2fsck_t ctx)
 {
 	ext2_filsys fs = ctx->fs;
-	int		i;
+	struct dir_info_iter *iter;
 #ifdef RESOURCE_TRACK
 	struct resource_track	rtrack;
 #endif
@@ -61,7 +61,7 @@ void e2fsck_pass3(e2fsck_t ctx)
 	unsigned long maxdirs, count;
 
 #ifdef RESOURCE_TRACK
-	init_resource_track(&rtrack);
+	init_resource_track(&rtrack, ctx->fs->io);
 #endif
 
 	clear_problem_context(&pctx);
@@ -87,7 +87,8 @@ void e2fsck_pass3(e2fsck_t ctx)
 #ifdef RESOURCE_TRACK
 	if (ctx->options & E2F_OPT_TIME) {
 		e2fsck_clear_progbar(ctx);
-		print_resource_track(_("Peak memory"), &ctx->global_rtrack);
+		print_resource_track(_("Peak memory"), &ctx->global_rtrack,
+				     NULL);
 	}
 #endif
 
@@ -103,16 +104,18 @@ void e2fsck_pass3(e2fsck_t ctx)
 	if (ctx->progress)
 		if ((ctx->progress)(ctx, 3, 0, maxdirs))
 			goto abort_exit;
-	
-	for (i=0; (dir = e2fsck_dir_info_iter(ctx, &i)) != 0;) {
+
+	iter = e2fsck_dir_info_iter_begin(ctx);
+	while ((dir = e2fsck_dir_info_iter(ctx, iter)) != 0) {
 		if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
 			goto abort_exit;
 		if (ctx->progress && (ctx->progress)(ctx, 3, count++, maxdirs))
 			goto abort_exit;
 		if (ext2fs_test_inode_bitmap(ctx->inode_dir_map, dir->ino))
-			if (check_directory(ctx, dir, &pctx))
+			if (check_directory(ctx, dir->ino, &pctx))
 				goto abort_exit;
 	}
+	e2fsck_dir_info_iter_end(ctx, iter);
 
 	/*
 	 * Force the creation of /lost+found if not present
@@ -125,7 +128,7 @@ void e2fsck_pass3(e2fsck_t ctx)
 	 * optimized, do it here.
 	 */
 	e2fsck_rehash_directories(ctx);
-	
+
 abort_exit:
 	e2fsck_free_dir_info(ctx);
 	if (inode_loop_detect) {
@@ -140,7 +143,7 @@ abort_exit:
 #ifdef RESOURCE_TRACK
 	if (ctx->options & E2F_OPT_TIME2) {
 		e2fsck_clear_progbar(ctx);
-		print_resource_track(_("Pass 3"), &rtrack);
+		print_resource_track(_("Pass 3"), &rtrack, ctx->fs->io);
 	}
 #endif
 }
@@ -156,9 +159,9 @@ static void check_root(e2fsck_t ctx)
 	struct ext2_inode	inode;
 	char *			block;
 	struct problem_context	pctx;
-	
+
 	clear_problem_context(&pctx);
-	
+
 	if (ext2fs_test_inode_bitmap(ctx->inode_used_map, EXT2_ROOT_INO)) {
 		/*
 		 * If the root inode is not a directory, die here.  The
@@ -180,7 +183,7 @@ static void check_root(e2fsck_t ctx)
 	}
 
 	e2fsck_read_bitmaps(ctx);
-	
+
 	/*
 	 * First, find a free block
 	 */
@@ -222,22 +225,22 @@ static void check_root(e2fsck_t ctx)
 	memset(&inode, 0, sizeof(inode));
 	inode.i_mode = 040755;
 	inode.i_size = fs->blocksize;
-	inode.i_atime = inode.i_ctime = inode.i_mtime = time(0);
+	inode.i_atime = inode.i_ctime = inode.i_mtime = ctx->now;
 	inode.i_links_count = 2;
-	inode.i_blocks = fs->blocksize / 512;
+	ext2fs_iblk_set(fs, &inode, 1);
 	inode.i_block[0] = blk;
 
 	/*
 	 * Write out the inode.
 	 */
-	pctx.errcode = ext2fs_write_inode(fs, EXT2_ROOT_INO, &inode);
+	pctx.errcode = ext2fs_write_new_inode(fs, EXT2_ROOT_INO, &inode);
 	if (pctx.errcode) {
 		pctx.str = "ext2fs_write_inode";
 		fix_problem(ctx, PR_3_CREATE_ROOT_ERROR, &pctx);
 		ctx->flags |= E2F_FLAG_ABORT;
 		return;
 	}
-	
+
 	/*
 	 * Miscellaneous bookkeeping...
 	 */
@@ -258,7 +261,7 @@ static void check_root(e2fsck_t ctx)
  * the lost+found.  We have to do loop detection; if we ever discover
  * a loop, we treat that as a disconnected directory and offer to
  * reparent it to lost+found.
- * 
+ *
  * However, loop detection is expensive, because for very large
  * filesystems, the inode_loop_detect bitmap is huge, and clearing it
  * is non-trivial.  Loops in filesystems are also a rare error case,
@@ -267,16 +270,13 @@ static void check_root(e2fsck_t ctx)
  * increment a counter; if the counter exceeds some extreme threshold,
  * then we try again with the loop detection bitmap enabled.
  */
-static int check_directory(e2fsck_t ctx, struct dir_info *dir,
+static int check_directory(e2fsck_t ctx, ext2_ino_t dir,
 			   struct problem_context *pctx)
 {
 	ext2_filsys 	fs = ctx->fs;
-	struct dir_info *p = dir;
+	ext2_ino_t	ino = dir, parent;
 	int		loop_pass = 0, parent_count = 0;
 
-	if (!p)
-		return 0;
-
 	while (1) {
 		/*
 		 * Mark this inode as being "done"; by the time we
@@ -288,38 +288,38 @@ static int check_directory(e2fsck_t ctx, struct dir_info *dir,
 		 * If it was marked done already, then we've reached a
 		 * parent we've already checked.
 		 */
-	  	if (ext2fs_mark_inode_bitmap(inode_done_map, p->ino))
+	  	if (ext2fs_mark_inode_bitmap(inode_done_map, ino))
 			break;
 
+		if (e2fsck_dir_info_get_parent(ctx, ino, &parent)) {
+			fix_problem(ctx, PR_3_NO_DIRINFO, pctx);
+			return 0;
+		}
+
 		/*
 		 * If this directory doesn't have a parent, or we've
 		 * seen the parent once already, then offer to
 		 * reparent it to lost+found
 		 */
-		if (!p->parent ||
-		    (loop_pass && 
+		if (!parent ||
+		    (loop_pass &&
 		     (ext2fs_test_inode_bitmap(inode_loop_detect,
-					      p->parent)))) {
-			pctx->ino = p->ino;
+					       parent)))) {
+			pctx->ino = ino;
 			if (fix_problem(ctx, PR_3_UNCONNECTED_DIR, pctx)) {
 				if (e2fsck_reconnect_file(ctx, pctx->ino))
 					ext2fs_unmark_valid(fs);
 				else {
-					p = e2fsck_get_dir_info(ctx, pctx->ino);
-					p->parent = ctx->lost_and_found;
-					fix_dotdot(ctx, p, ctx->lost_and_found);
+					fix_dotdot(ctx, pctx->ino,
+						   ctx->lost_and_found);
+					parent = ctx->lost_and_found;
 				}
 			}
 			break;
 		}
-		p = e2fsck_get_dir_info(ctx, p->parent);
-		if (!p) {
-			fix_problem(ctx, PR_3_NO_DIRINFO, pctx);
-			return 0;
-		}
+		ino = parent;
 		if (loop_pass) {
-			ext2fs_mark_inode_bitmap(inode_loop_detect,
-						 p->ino);
+			ext2fs_mark_inode_bitmap(inode_loop_detect, ino);
 		} else if (parent_count++ > 2048) {
 			/*
 			 * If we've run into a path depth that's
@@ -334,13 +334,13 @@ static int check_directory(e2fsck_t ctx, struct dir_info *dir,
 				pctx->errcode = ext2fs_allocate_inode_bitmap(fs, _("inode loop detection bitmap"), &inode_loop_detect);
 				if (pctx->errcode) {
 					pctx->num = 1;
-					fix_problem(ctx, 
+					fix_problem(ctx,
 				    PR_3_ALLOCATE_IBITMAP_ERROR, pctx);
 					ctx->flags |= E2F_FLAG_ABORT;
 					return -1;
 				}
 			}
-			p = dir;
+			ino = dir;
 		}
 	}
 
@@ -348,12 +348,15 @@ static int check_directory(e2fsck_t ctx, struct dir_info *dir,
 	 * Make sure that .. and the parent directory are the same;
 	 * offer to fix it if not.
 	 */
-	if (dir->parent != dir->dotdot) {
-		pctx->ino = dir->ino;
-		pctx->ino2 = dir->dotdot;
-		pctx->dir = dir->parent;
+	pctx->ino = dir;
+	if (e2fsck_dir_info_get_dotdot(ctx, dir, &pctx->ino2) ||
+	    e2fsck_dir_info_get_parent(ctx, dir, &pctx->dir)) {
+		fix_problem(ctx, PR_3_NO_DIRINFO, pctx);
+		return 0;
+	}
+	if (pctx->ino2 != pctx->dir) {
 		if (fix_problem(ctx, PR_3_BAD_DOT_DOT, pctx))
-			fix_dotdot(ctx, dir, dir->parent);
+			fix_dotdot(ctx, dir, pctx->dir);
 	}
 	return 0;
 }
@@ -372,13 +375,12 @@ ext2_ino_t e2fsck_get_lost_and_found(e2fsck_t ctx, int fix)
 	char *			block;
 	static const char	name[] = "lost+found";
 	struct 	problem_context	pctx;
-	struct dir_info 	*dirinfo;
 
 	if (ctx->lost_and_found)
 		return ctx->lost_and_found;
 
 	clear_problem_context(&pctx);
-	
+
 	retval = ext2fs_lookup(fs, EXT2_ROOT_INO, name,
 			       sizeof(name)-1, 0, &ino);
 	if (retval && !fix)
@@ -388,7 +390,7 @@ ext2_ino_t e2fsck_get_lost_and_found(e2fsck_t ctx, int fix)
 			ctx->lost_and_found = ino;
 			return ino;
 		}
-				
+
 		/* Lost+found isn't a directory! */
 		if (!fix)
 			return 0;
@@ -403,9 +405,7 @@ ext2_ino_t e2fsck_get_lost_and_found(e2fsck_t ctx, int fix)
 			fix_problem(ctx, PR_3_CREATE_LPF_ERROR, &pctx);
 			return 0;
 		}
-		dirinfo = e2fsck_get_dir_info(ctx, ino);
-		if (dirinfo)
-			dirinfo->parent = 0;
+		(void) e2fsck_dir_info_set_parent(ctx, ino, 0);
 		e2fsck_adjust_inode_count(ctx, ino, -1);
 	} else if (retval != EXT2_ET_FILE_NOT_FOUND) {
 		pctx.errcode = retval;
@@ -419,7 +419,7 @@ ext2_ino_t e2fsck_get_lost_and_found(e2fsck_t ctx, int fix)
 	 * them.
 	 */
 	e2fsck_read_bitmaps(ctx);
-	
+
 	/*
 	 * First, find a free block
 	 */
@@ -470,15 +470,15 @@ ext2_ino_t e2fsck_get_lost_and_found(e2fsck_t ctx, int fix)
 	memset(&inode, 0, sizeof(inode));
 	inode.i_mode = 040700;
 	inode.i_size = fs->blocksize;
-	inode.i_atime = inode.i_ctime = inode.i_mtime = time(0);
+	inode.i_atime = inode.i_ctime = inode.i_mtime = ctx->now;
 	inode.i_links_count = 2;
-	inode.i_blocks = fs->blocksize / 512;
+	ext2fs_iblk_set(fs, &inode, 1);
 	inode.i_block[0] = blk;
 
 	/*
 	 * Next, write out the inode.
 	 */
-	pctx.errcode = ext2fs_write_inode(fs, ino, &inode);
+	pctx.errcode = ext2fs_write_new_inode(fs, ino, &inode);
 	if (pctx.errcode) {
 		pctx.str = "ext2fs_write_inode";
 		fix_problem(ctx, PR_3_CREATE_LPF_ERROR, &pctx);
@@ -531,7 +531,7 @@ int e2fsck_reconnect_file(e2fsck_t ctx, ext2_ino_t ino)
 		fix_problem(ctx, PR_3_NO_LPF, &pctx);
 		return 1;
 	}
-	
+
 	sprintf(name, "#%u", ino);
 	if (ext2fs_read_inode(fs, ino, &inode) == 0)
 		file_type = ext2_file_type(inode.i_mode);
@@ -539,7 +539,7 @@ int e2fsck_reconnect_file(e2fsck_t ctx, ext2_ino_t ino)
 	if (retval == EXT2_ET_DIR_NO_SPACE) {
 		if (!fix_problem(ctx, PR_3_EXPAND_LF_DIR, &pctx))
 			return 1;
-		retval = e2fsck_expand_directory(ctx, ctx->lost_and_found, 
+		retval = e2fsck_expand_directory(ctx, ctx->lost_and_found,
 						 1, 0);
 		if (retval) {
 			pctx.errcode = retval;
@@ -567,7 +567,7 @@ errcode_t e2fsck_adjust_inode_count(e2fsck_t ctx, ext2_ino_t ino, int adj)
 	ext2_filsys fs = ctx->fs;
 	errcode_t		retval;
 	struct ext2_inode 	inode;
-	
+
 	if (!ino)
 		return 0;
 
@@ -593,7 +593,7 @@ errcode_t e2fsck_adjust_inode_count(e2fsck_t ctx, ext2_ino_t ino, int adj)
 		ext2fs_icount_decrement(ctx->inode_link_info, ino, 0);
 		inode.i_links_count--;
 	}
-	
+
 	retval = ext2fs_write_inode(fs, ino, &inode);
 	if (retval)
 		return retval;
@@ -627,7 +627,7 @@ static int fix_dotdot_proc(struct ext2_dir_entry *dirent,
 		return 0;
 
 	clear_problem_context(&pctx);
-	
+
 	retval = e2fsck_adjust_inode_count(fp->ctx, dirent->inode, -1);
 	if (retval) {
 		pctx.errcode = retval;
@@ -639,12 +639,18 @@ static int fix_dotdot_proc(struct ext2_dir_entry *dirent,
 		fix_problem(fp->ctx, PR_3_ADJUST_INODE, &pctx);
 	}
 	dirent->inode = fp->parent;
+	if (fp->ctx->fs->super->s_feature_incompat &
+	    EXT2_FEATURE_INCOMPAT_FILETYPE)
+		dirent->name_len = (dirent->name_len & 0xFF) |
+			(EXT2_FT_DIR << 8);
+	else
+		dirent->name_len = dirent->name_len & 0xFF;
 
 	fp->done++;
 	return DIRENT_ABORT | DIRENT_CHANGED;
 }
 
-static void fix_dotdot(e2fsck_t ctx, struct dir_info *dir, ext2_ino_t parent)
+static void fix_dotdot(e2fsck_t ctx, ext2_ino_t ino, ext2_ino_t parent)
 {
 	ext2_filsys fs = ctx->fs;
 	errcode_t	retval;
@@ -657,21 +663,23 @@ static void fix_dotdot(e2fsck_t ctx, struct dir_info *dir, ext2_ino_t parent)
 	fp.ctx = ctx;
 
 #if 0
-	printf("Fixing '..' of inode %lu to be %lu...\n", dir->ino, parent);
+	printf("Fixing '..' of inode %lu to be %lu...\n", ino, parent);
 #endif
-	
-	retval = ext2fs_dir_iterate(fs, dir->ino, DIRENT_FLAG_INCLUDE_EMPTY,
+
+	clear_problem_context(&pctx);
+	pctx.ino = ino;
+	retval = ext2fs_dir_iterate(fs, ino, DIRENT_FLAG_INCLUDE_EMPTY,
 				    0, fix_dotdot_proc, &fp);
 	if (retval || !fp.done) {
-		clear_problem_context(&pctx);
-		pctx.ino = dir->ino;
 		pctx.errcode = retval;
 		fix_problem(ctx, retval ? PR_3_FIX_PARENT_ERR :
 			    PR_3_FIX_PARENT_NOFIND, &pctx);
 		ext2fs_unmark_valid(fs);
 	}
-	dir->dotdot = parent;
-	
+	(void) e2fsck_dir_info_set_dotdot(ctx, ino, parent);
+	if (e2fsck_dir_info_set_parent(ctx, ino, ctx->lost_and_found))
+		fix_problem(ctx, PR_3_NO_DIRINFO, &pctx);
+
 	return;
 }
 
@@ -704,7 +712,7 @@ static int expand_dir_proc(ext2_filsys fs,
 	e2fsck_t	ctx;
 
 	ctx = es->ctx;
-	
+
 	if (es->guaranteed_size && blockcnt >= es->guaranteed_size)
 		return BLOCK_ABORT;
 
@@ -736,7 +744,7 @@ static int expand_dir_proc(ext2_filsys fs,
 		}
 		memset(block, 0, fs->blocksize);
 		retval = io_channel_write_blk(fs->io, new_blk, 1, block);
-	}	
+	}
 	if (retval) {
 		es->err = retval;
 		return BLOCK_ABORT;
@@ -746,7 +754,7 @@ static int expand_dir_proc(ext2_filsys fs,
 	ext2fs_mark_block_bitmap(ctx->block_found_map, new_blk);
 	ext2fs_block_alloc_stats(fs, new_blk, +1);
 	es->newblocks++;
-	
+
 	if (es->num == 0)
 		return (BLOCK_CHANGED | BLOCK_ABORT);
 	else
@@ -760,7 +768,7 @@ errcode_t e2fsck_expand_directory(e2fsck_t ctx, ext2_ino_t dir,
 	errcode_t	retval;
 	struct expand_dir_struct es;
 	struct ext2_inode	inode;
-	
+
 	if (!(fs->flags & EXT2_FLAG_RW))
 		return EXT2_ET_RO_FILSYS;
 
@@ -773,14 +781,14 @@ errcode_t e2fsck_expand_directory(e2fsck_t ctx, ext2_ino_t dir,
 	retval = ext2fs_check_directory(fs, dir);
 	if (retval)
 		return retval;
-	
+
 	es.num = num;
 	es.guaranteed_size = guaranteed_size;
 	es.last_block = 0;
 	es.err = 0;
 	es.newblocks = 0;
 	es.ctx = ctx;
-	
+
 	retval = ext2fs_block_iterate2(fs, dir, BLOCK_FLAG_APPEND,
 				       0, expand_dir_proc, &es);
 
@@ -793,9 +801,9 @@ errcode_t e2fsck_expand_directory(e2fsck_t ctx, ext2_ino_t dir,
 	retval = ext2fs_read_inode(fs, dir, &inode);
 	if (retval)
 		return retval;
-	
+
 	inode.i_size = (es.last_block + 1) * fs->blocksize;
-	inode.i_blocks += (fs->blocksize / 512) * es.newblocks;
+	ext2fs_iblk_add_blocks(fs, &inode, es.newblocks);
 
 	e2fsck_write_inode(ctx, dir, &inode, "expand_directory");
 
