@@ -10,6 +10,8 @@
  * %End-Header%
  */
 
+#define _XOPEN_SOURCE 600 /* for inclusion of sa_handler in Solaris */
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -19,6 +21,9 @@
 #include <fcntl.h>
 #include <time.h>
 #include <errno.h>
+#ifdef HAVE_SIGNAL_H
+#include <signal.h>
+#endif
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
 #else
@@ -32,6 +37,7 @@ void	*outbuf = 0;
 int	verbose = 0;
 int	do_skip = 0;
 int	skip_mode = 0;
+pid_t	child_pid = -1;
 
 static void usage(char *progname)
 {
@@ -46,10 +52,10 @@ static void usage(char *progname)
 static void send_output(const char *buffer, int c, int flag)
 {
 	char	*n;
-	
+
 	if (c == 0)
 		c = strlen(buffer);
-	
+
 	if (flag & SEND_CONSOLE)
 		write(1, buffer, c);
 	if (!(flag & SEND_LOG))
@@ -102,17 +108,36 @@ static int do_read(int fd)
 	return c;
 }
 
+static void signal_term(int sig)
+{
+	if (child_pid > 0)
+		kill(child_pid, sig);
+}
+
 static int run_program(char **argv)
 {
 	int	fds[2];
 	int	status, rc, pid;
 	char	buffer[80];
+#ifdef HAVE_SIGNAL_H
+	struct sigaction	sa;
+#endif
 
 	if (pipe(fds) < 0) {
 		perror("pipe");
 		exit(1);
 	}
 
+#ifdef HAVE_SIGNAL_H
+	memset(&sa, 0, sizeof(struct sigaction));
+	sa.sa_handler = signal_term;
+	sigaction(SIGINT, &sa, 0);
+	sigaction(SIGTERM, &sa, 0);
+#ifdef SA_RESTART
+	sa.sa_flags = SA_RESTART;
+#endif
+#endif
+
 	pid = fork();
 	if (pid < 0) {
 		perror("vfork");
@@ -122,16 +147,18 @@ static int run_program(char **argv)
 		dup2(fds[1],1);		/* fds[1] replaces stdout */
 		dup2(fds[1],2);  	/* fds[1] replaces stderr */
 		close(fds[0]);	/* don't need this here */
-		
+
 		execvp(argv[0], argv);
 		perror(argv[0]);
 		exit(1);
 	}
+	child_pid = pid;
 	close(fds[1]);
 
 	while (!(waitpid(pid, &status, WNOHANG ))) {
 		do_read(fds[0]);
 	}
+	child_pid = -1;
 	do_read(fds[0]);
 	close(fds[0]);
 
@@ -174,8 +201,8 @@ static int copy_from_stdin(void)
 		bad_read = 0;
 	}
 	return 0;
-}	
-	
+}
+
 
 
 int main(int argc, char **argv)
@@ -186,7 +213,7 @@ int main(int argc, char **argv)
 	int	send_flag = SEND_LOG;
 	int	do_stdin;
 	time_t	t;
-	
+
 	while ((c = getopt(argc, argv, "+asv")) != EOF) {
 		switch (c) {
 		case 'a':
@@ -211,7 +238,7 @@ int main(int argc, char **argv)
 
 	outfd = open(outfn, openflags, 0644);
 	do_stdin = !strcmp(argv[0], "-");
-	
+
 	send_output("Log of ", 0, send_flag);
 	if (do_stdin)
 		send_output("stdin", 0, send_flag);
@@ -230,7 +257,7 @@ int main(int argc, char **argv)
 		rc = copy_from_stdin();
 	else
 		rc = run_program(argv);
-	
+
 	send_output("\n", 0, send_flag);
 	t = time(0);
 	send_output(ctime(&t), 0, send_flag);
@@ -252,11 +279,11 @@ int main(int argc, char **argv)
 		while (outfd < 0) {
 			outfd = open(outfn, openflags, 0644);
 			sleep(1);
-		} 
+		}
 		write(outfd, outbuf, outbufsize);
 		free(outbuf);
 	}
 	close(outfd);
-	
+
 	exit(rc);
 }
